from active_bo_msgs.srv import ActiveBO
from active_bo_msgs.srv import ActiveRL

import rclpy
from rclpy.node import Node

from active_bo_ros.BayesianOptimization.BayesianOptimization import BayesianOptimization
from active_bo_ros.ReinforcementLearning.ContinuousMountainCar import Continuous_MountainCarEnv

import numpy as np


class ActiveBOService(Node):
    def __init__(self):
        super().__init__('active_bo_service')
        self.srv = self.create_service(ActiveBO, 'active_bo_srv', self.active_bo_callback)

        self.active_rl_client = self.create_client(ActiveRL, 'active_rl_srv')

        self.env = Continuous_MountainCarEnv()
        self.distance_penalty = 0

        self.nr_init = 3

    def active_bo_callback(self, request, response):
        self.get_logger().info('Active Bayesian Optimization Service started!')
        nr_weights = request.nr_weights
        max_steps = request.max_steps
        nr_episodes = request.nr_episodes
        nr_runs = request.nr_runs
        acq = request.acquisition_function
        epsilon = request.epsilon

        reward = np.zeros((nr_episodes, nr_runs))
        best_pol_reward = np.zeros((1, nr_runs))
        best_policy = np.zeros((max_steps, nr_runs))
        best_weights = np.zeros((nr_weights, nr_runs))

        BO = BayesianOptimization(self.env,
                                  max_steps,
                                  nr_init=self.nr_init,
                                  acq=acq,
                                  nr_weights=nr_weights)

        arl_request = ActiveRL.Request()
        for i in range(nr_runs):
            BO.initialize()

            for j in range(nr_episodes):
                # active part
                if (j > 0) and (np.random.uniform(0.0, 1.0, 1) < epsilon):
                    self.get_logger().info('Active User Input')
                    old_policy, _, old_weights = BO.get_best_result()

                    arl_request.old_policy = old_policy.tolist()
                    arl_request.old_weights = old_weights.tolist()
                    arl_response = self.active_rl_client.call(arl_request)

                    BO.add_new_observation(arl_response.reward, arl_response.new_weights)

                # BO part
                else:
                    x_next = BO.next_observation()
                    BO.eval_new_observation(x_next)

                self.get_logger().info(str(j))

            best_policy[:, i], best_pol_reward[:, i], best_weights[:, i] = BO.get_best_result()

            reward[:, i] = BO.best_reward.T

        response.reward_mean = np.mean(reward, axis=1).tolist()
        response.reward_std = np.std(reward, axis=1).tolist()

        best_policy_idx = np.argmax(best_pol_reward)
        response.best_weights = best_weights[:, best_policy_idx].tolist()
        response.best_policy = best_policy[:, best_policy_idx].tolist()
        return response


def main(args=None):
    rclpy.init(args=args)

    active_bo_service = ActiveBOService()

    rclpy.spin(active_bo_service)


if __name__ == '__main__':
    main()
